#  Copyright (c) <2014> <thenoviceoof>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

import ast
from pyli.refs import find_free_references
from pyli.preamble import create_imports
from pyli.spec import handle_special_variables
from pyli.util import var_base_difference, var_base_intersection
import logging
import sys

LOG = logging.getLogger(__name__)


def main(
    code: str,
    debug: int = logging.ERROR,
    pprint_opt: bool = False,
    variables: dict = {},
) -> None:
    # Set logging verbosity.
    logging.basicConfig(level=debug)

    # Parse the code.
    tree = ast.parse(code)
    LOG.debug("Initial parse tree...")
    LOG.debug(ast.dump(tree, indent=4))

    # Find the free variables.
    free_vars = find_free_references(tree)
    LOG.debug("Free variables found: {}".format(free_vars))
    if pprint_opt:
        free_vars.add(("pprint",))

    if debug != logging.ERROR and var_base_intersection(free_vars, {"stderr"}):
        LOG.error("Conflictng use of debug logging and writing to stderr.")
        sys.exit(2)

    # Handle any special variables and output on a case-by-case basis.
    free_vars = handle_special_variables(tree, free_vars, pprint_opt)
    # We will pass in command line variables via exec.
    free_vars = var_base_difference(free_vars, {k for k in variables.keys()})

    # Add imports for the rest of the free variables.
    create_imports(tree, free_vars)

    # Compile and execute the code.
    ast.fix_missing_locations(tree)
    LOG.debug("Final parse tree...")
    LOG.debug(ast.dump(tree, indent=4))
    LOG.info("Compiling and executing code...")
    bytecode = compile(
        tree,
        "<generated code>",  # "filename", used in tracebacks
        "exec",  # Mode, multiple statements (instead of expr)
    )
    # Create a clean context, since test cases might leak the default
    # arg dict across runs.
    context = dict(**variables)
    # Since we're executing inside of main(), any imports are actually
    # locals. Providing a globals dict prevents leaking any dev
    # environment leaks, and is used as a locals, meaning that any
    # "local" imports end up in the "globals" namespace.
    # See https://stackoverflow.com/a/12505166
    exec(
        bytecode,
        context,  # Globals
        # If not locals dict is given, globals=locals.
    )



#  Copyright (c) <2014> <thenoviceoof>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

import ast
import logging
import sys
from collections.abc import Sequence
from pyli.util import var_base_intersection, var_base_difference

LOG = logging.getLogger(__name__)

PREFIX = "PYLI_RESERVED_"

SPEC_PER_LINE = {"l", "li", "line"}
SPEC_LINE_GEN = {"ls", "lis", "lines"}
SPEC_CONTENTS = {"cs", "conts", "contents"}
SPEC_PER_PART = {"p", "part"}
SPEC_PARTS_GEN = {"ps", "parts"}
SPEC_STD = {"stdin", "stdout", "stderr"}


def handle_special_variables(
    tree: ast.Module, free_variables: set[tuple[str, ...]], pprint: bool
) -> set[tuple[str, ...]]:
    LOG.info("Handling special variables...")
    if var_base_intersection(free_variables, SPEC_PER_LINE):
        LOG.debug("Per-line variables detected")
        # Create a stdin line generator.
        stdin_nodes = create_stdin_reader_lines()
        tmp_line_name = PREFIX + "line"
        aliasing = [
            set_variable_to_name(v, tmp_line_name)
            for v in var_base_intersection(free_variables, SPEC_PER_LINE)
        ]
        wrap_last_statement_with_print(tree.body, pprint)
        ast.increment_lineno(tree, stdin_nodes[-1].lineno)
        # Execute the code per line.
        for_node = ast.For(
            target=ast.Name(id=tmp_line_name, ctx=ast.Store()),
            iter=ast.Name(id=PREFIX + "lines", ctx=ast.Load()),
            body=aliasing + tree.body,
            orelse=[],
        )
        tree.body = stdin_nodes + [for_node]
        return var_base_difference(free_variables, SPEC_PER_LINE) | {("sys",)}
    elif var_base_intersection(free_variables, SPEC_LINE_GEN):
        LOG.debug("Line generator variables detected")
        # Create a stdin line generator.
        stdin_nodes = create_stdin_reader_lines()
        aliasing = [
            set_variable_to_name(v, PREFIX + "lines")
            for v in var_base_intersection(free_variables, SPEC_LINE_GEN)
        ]
        # Wrap the last statement with print(...).
        wrap_last_statement_with_print(tree.body, pprint)
        ast.increment_lineno(tree, stdin_nodes[-1].lineno + len(aliasing))
        tree.body = stdin_nodes + aliasing + tree.body
        return var_base_difference(free_variables, SPEC_LINE_GEN) | {("sys",)}
    elif var_base_intersection(free_variables, SPEC_CONTENTS):
        LOG.debug("Contents variables detected")
        # Create a stdin line generator.
        stdin_node = ast.Assign(
            targets=[ast.Name(id=PREFIX + "contents", ctx=ast.Store())],
            value=ast.Call(
                func=ast_attr(("sys", "stdin", "read")), args=[], keywords=[]
            ),
        )
        aliasing = [
            set_variable_to_name(v, PREFIX + "contents")
            for v in var_base_intersection(free_variables, SPEC_CONTENTS)
        ]
        # Wrap the last statement with print(...).
        wrap_last_statement_with_print(tree.body, pprint)
        ast.increment_lineno(tree, 1 + len(aliasing))
        tree.body = [stdin_node] + aliasing + tree.body
        return var_base_difference(free_variables, SPEC_CONTENTS) | {("sys",)}
    elif var_base_intersection(free_variables, SPEC_PER_PART):
        LOG.debug("Space-delimited parts variables detected")
        # Create a stdin space-delimited parts generator.
        stdin_nodes = create_stdin_reader_parts()
        # Wrap the last statement with print(...).
        wrap_last_statement_with_print(tree.body, pprint)
        aliasing = [
            set_variable_to_name(v, PREFIX + "part")
            for v in var_base_intersection(free_variables, SPEC_PER_PART)
        ]
        ast.increment_lineno(tree, 1 + stdin_nodes[-1].lineno + len(aliasing))
        for_node = ast.For(
            target=ast.Name(id=PREFIX + "part", ctx=ast.Store()),
            iter=ast.Name(id=PREFIX + "parts", ctx=ast.Load()),
            body=aliasing + tree.body,
            orelse=[],
        )
        tree.body = stdin_nodes + [for_node]
        return var_base_difference(free_variables, SPEC_PER_PART) | {("sys",)}
    elif var_base_intersection(free_variables, SPEC_PARTS_GEN):
        LOG.debug("Space-delimited line generator detected")
        # Create a stdin space-delimited parts generator.
        stdin_nodes = create_stdin_reader_parts()
        # Wrap the last statement with print(...).
        wrap_last_statement_with_print(tree.body, pprint)
        aliasing = [
            set_variable_to_name(v, PREFIX + "parts")
            for v in var_base_intersection(free_variables, SPEC_PARTS_GEN)
        ]
        ast.increment_lineno(tree, stdin_nodes[-1].lineno + len(aliasing))
        tree.body = stdin_nodes + aliasing + tree.body
        return var_base_difference(free_variables, SPEC_PARTS_GEN) | {("sys",)}
    elif var_base_intersection(free_variables, SPEC_STD):
        LOG.debug("std* reference detected")
        aliasing = []
        if var_base_intersection(free_variables, {"stdin"}):
            aliasing.append(set_variable_to_node("stdin", ast_attr(["sys", "stdin"])))
        if var_base_intersection(free_variables, {"stdout"}):
            aliasing.append(set_variable_to_node("stdout", ast_attr(["sys", "stdout"])))
        if var_base_intersection(free_variables, {"stderr"}):
            aliasing.append(set_variable_to_node("stderr", ast_attr(["sys", "stderr"])))
        for alias in aliasing:
            ast.copy_location(alias, tree.body[0])
        # If you're using stdout, you probably want only specific things going to stdout.
        if not var_base_intersection(free_variables, {"stdout"}):
            wrap_last_statement_with_print(tree.body, pprint)
        ast.increment_lineno(tree, 1 + len(aliasing))
        tree.body = aliasing + tree.body
        return var_base_difference(free_variables, SPEC_STD) | {("sys",)}
    else:
        # No special behavior required, just make sure to print the last statement.
        LOG.debug("No special variable behavior detected")
        wrap_last_statement_with_print(tree.body, pprint)
        return free_variables


def set_variable_to_node(target_name: str, source_node: ast.expr) -> ast.Assign:
    return ast.Assign(
        targets=[ast.Name(id=target_name, ctx=ast.Store())], value=source_node
    )


def set_variable_to_name(target_name: str, source_name: str) -> ast.Assign:
    return set_variable_to_node(target_name, ast.Name(id=source_name, ctx=ast.Load()))


def ast_attr(parts: Sequence[str], load: bool = True) -> ast.expr:
    """Produce an AST representing a dot access chain."""
    if len(parts) == 1:
        return ast.Name(id=parts[0], ctx=ast.Load() if load else ast.Store())
    else:
        return ast.Attribute(
            value=ast_attr(parts[:-1], load),
            attr=parts[-1],
            ctx=ast.Load() if load else ast.Store(),
        )


def create_print_ast(
    node: ast.expr, pprint: bool, location_node: ast.AST
) -> list[ast.stmt]:
    """Wraps a given AST node into a print pattern; don't print None."""
    # Assign the result of the node to a variable, since the node could have side effects.
    tmp_var_name = PREFIX + "tmp_print_holder"
    assign_node = ast.Assign(
        targets=[ast.Name(id=tmp_var_name, ctx=ast.Store())], value=node
    )
    print_expr = (
        ast.Call(
            func=ast.Attribute(
                value=ast.Name(id="pprint", ctx=ast.Load()),
                attr="pprint",
                ctx=ast.Load(),
            ),
            args=[ast.Name(id=tmp_var_name, ctx=ast.Load())],
            keywords=[],
        )
        if pprint
        else ast.Call(
            func=ast.Name(id="print", ctx=ast.Load()),
            args=[ast.Name(id=tmp_var_name, ctx=ast.Load())],
            keywords=[],
        )
    )
    none_check = ast.If(
        test=ast.Compare(
            left=ast.Name(id=tmp_var_name, ctx=ast.Load()),
            ops=[ast.IsNot()],
            comparators=[ast.Constant(value=None)],
        ),
        body=[ast.Expr(value=print_expr)],
        orelse=[],
    )
    ast.copy_location(assign_node, location_node)
    ast.copy_location(none_check, location_node)
    return [assign_node, none_check]


def wrap_last_statement_with_print(stmts: list[ast.stmt], pprint: bool) -> None:
    """Given an AST body, wrap the "last" statement in a call to print(...)."""
    last_node = stmts[-1]
    if isinstance(last_node, ast.Expr):
        # Check if the expression is already wrapped in a print statement.
        if is_ast_print(last_node.value):
            return
        # If our given node is an Expr (statement-expression), unwrap
        # it before putting it into a print(...).
        # See https://stackoverflow.com/a/32429203
        print_stmts = create_print_ast(last_node.value, pprint, last_node)
        stmts.pop()
        stmts.extend(print_stmts)
    elif (
        isinstance(last_node, ast.Assign)
        or isinstance(last_node, ast.AnnAssign)
        or isinstance(last_node, ast.AugAssign)
    ):
        # If there's no actual assignment, then there's nothing to print.
        if not hasattr(last_node, "value"):
            return
        # - Why don't we just transform `x = 1` into `print(1)`?
        #   `x.a = 1` could potentially do weird things, by overriding __setattr__.
        #   Additionally, the rhs might be non-idempotent.
        #   Therefore, we only print the last reference.
        # - If there are multiple targets, we are parsing something like
        #   `a = b = 1`, and only need the first reference.
        # - If there is destructuring, we will get a ast.Tuple object to print.
        target = None
        if isinstance(last_node, ast.Assign):
            target = last_node.targets[0]
        else:
            target = last_node.target
        target = set_assignment_target_context(target, ast.Load())
        print_stmts = create_print_ast(target, pprint, last_node)
        stmts.extend(print_stmts)
    elif isinstance(last_node, ast.If):
        # That is disgusting, elif is represented extra If nodes.
        # Either way, we need to print the last statement of each
        # branch, and since we recurse we should get every branch
        # automatically.
        wrap_last_statement_with_print(last_node.body, pprint)
        if len(last_node.orelse):
            wrap_last_statement_with_print(last_node.orelse, pprint)
    elif isinstance(last_node, ast.For) or isinstance(last_node, ast.While):
        if last_node.orelse:
            wrap_last_statement_with_print(last_node.orelse, pprint)
        else:
            wrap_last_statement_with_print(last_node.body, pprint)
    elif isinstance(last_node, ast.Try):
        if last_node.finalbody:
            wrap_last_statement_with_print(last_node.finalbody, pprint)
            # Finally always executes, so it is the definitive last statement.
            return
        if last_node.orelse:
            # Else executes if no exceptions are handled.
            wrap_last_statement_with_print(last_node.orelse, pprint)
        else:
            wrap_last_statement_with_print(last_node.body, pprint)
        # I think you could argue that we should not be printing
        # exception handlers, but the user is doing custom work they
        # might want printed.
        for handler in last_node.handlers:
            assert isinstance(handler, ast.ExceptHandler)
            wrap_last_statement_with_print(handler.body, pprint)
    elif isinstance(last_node, ast.With):
        wrap_last_statement_with_print(last_node.body, pprint)
    elif sys.version_info >= (3, 10, 0) and isinstance(last_node, ast.Match):
        for case in last_node.cases:
            assert isinstance(case, ast.match_case)
            wrap_last_statement_with_print(case.body, pprint)
    else:
        # There are many statements that shouldn't be printed, like
        # Raise or Assert, enough that I will not bother explicitly
        # listing them out.
        # In days of yore, we used to print function/class
        # definitions, but upon reflection who would want to see that?
        pass


def is_ast_print(node: ast.AST) -> bool:
    """Check if the given node represents a print(...) call."""
    return (
        isinstance(node, ast.Call)
        and isinstance(node.func, ast.Name)
        and node.func.id == "print"
    )


def create_stdin_reader_lines() -> list[ast.stmt]:
    code = """
def {fn}():
    while True:
        li = sys.stdin.readline()
        if not li:
            break
        yield li.rstrip('\\n')
{gen} = {fn}()
    """.format(
        fn=PREFIX + "line_generator", gen=PREFIX + "lines"
    )
    tmp_tree = ast.parse(code)
    return tmp_tree.body


def create_stdin_reader_parts() -> list[ast.stmt]:
    code = """
def {fn}():
    while True:
        li = sys.stdin.readline()
        if not li:
            break
        yield li.rstrip('\\n').split(' ')
{gen} = {fn}()
    """.format(
        fn=PREFIX + "parts_generator", gen=PREFIX + "parts"
    )
    tmp_tree = ast.parse(code)
    return tmp_tree.body


def set_assignment_target_context(
    target: ast.expr, context: ast.expr_context
) -> ast.expr:
    """Iterate through an assignment target and switch between
    load/store context.
    """
    # Thankfully, there is a limited number of ast nodes involved in assignment.
    if isinstance(target, ast.Attribute):
        return ast.Attribute(
            set_assignment_target_context(target.value, context), target.attr, context
        )
    elif isinstance(target, ast.Subscript):
        return ast.Subscript(
            set_assignment_target_context(target.value, context),
            set_assignment_target_context(target.slice, context),
            context,
        )
    elif isinstance(target, ast.Starred):
        return ast.Starred(
            set_assignment_target_context(target.value, context), context
        )
    elif isinstance(target, ast.Name):
        return ast.Name(target.id, context)
    elif isinstance(target, ast.List):
        return ast.List(
            [set_assignment_target_context(elt, context) for elt in target.elts],
            context,
        )
    elif isinstance(target, ast.Tuple):
        return ast.Tuple(
            [set_assignment_target_context(elt, context) for elt in target.elts],
            context,
        )
    else:
        raise ValueError(
            "Unexpected node type found in assignment context: {}".format(target)
        )



#  Copyright (c) <2014> <thenoviceoof>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

import ast
import builtins
import logging
import sys
from collections.abc import Sequence
from pyli.util import var_base_difference

LOG = logging.getLogger(__name__)

# There are 2 possible approaches to doing reference handling. Consider this example:
# `math.sqrt(2); math = 10`
# Should we insert `import math`, or not?
#   1. `math` is technically unbound at the time of first usage, and
#      this is valid python, so we should import the module. The
#      downside is that this becomes more complicated; what about
#      variables that are only set in one branch of an if-else
#      statement? Afterwards the variables could be treated as bound
#      or unbound, depending on runtime behavior. We could try to
#      import anyways, but this seems kind of messy.
#   2. Why the hell are we encouraging these shenanigans?  In English
#      we have 26 lowercase one-letter variable names, what do you
#      mean you want to re-use variables? The more likely cause of
#      these scenarios is probably incorrect input, and there's not
#      much we can do about that. Instead of trying to make
#      complicated rules, we can just opt for simpler rules, for a
#      simple tool.
#
# All that to say, if there is any binding anywhere, pyli will treat
# that variable as bound.
# Exceptions:
#  - list comprehension bindings are deliberately not accounted for
#    anywhere. List comprehensions are a nice shorthand, and we want
#    to be able to overload, say, `l`.

BUILTIN_NAMES = set(name for name in dir(builtins) if not name.startswith("_"))


def find_free_references(node: ast.AST) -> set[tuple[str, ...]]:
    """Recurse through an AST and find all the unbound references, without the builtins"""
    LOG.info("Looking for variables...")
    # Python keywords are consumed by parsing the ast, so there's no
    # need to worry about those.
    bound_vars, refs = find_references(node)
    LOG.debug("Found bound variables: {}".format(bound_vars))
    LOG.debug("Found references (including builtins): {}".format(refs))
    return var_base_difference(var_base_difference(refs, BUILTIN_NAMES), bound_vars)


def find_references(node: ast.AST) -> tuple[set[str], set[tuple[str, ...]]]:
    """Recurse through an AST and find all the bound variables and references."""
    # Start by recursing past everything that doesn't deal directly with variables.
    # Nodes are ordered as in
    # https://docs.python.org/3/library/ast.html, except if there's
    # some special processing necessary.
    if isinstance(node, ast.Module):
        return find_multiple_node_references(node.body)
    elif isinstance(node, ast.FormattedValue):
        return find_references(node.value)
    elif isinstance(node, ast.JoinedStr):
        return find_multiple_node_references(node.values)
    elif (
        isinstance(node, ast.List)
        or isinstance(node, ast.Tuple)
        or isinstance(node, ast.Set)
    ):
        return find_multiple_node_references(node.elts)
    elif isinstance(node, ast.Dict):
        # Possible to unpack a dictionary into a literal, which has a None key.
        # Example: `{'a': 1, **existing_dict}`
        return find_multiple_node_references(
            [key for key in node.keys if key is not None] + node.values
        )
    elif isinstance(node, ast.Expr):
        return find_references(node.value)
    elif isinstance(node, ast.UnaryOp):
        return find_references(node.operand)
    elif isinstance(node, ast.BinOp):
        return find_multiple_node_references([node.left, node.right])
    elif isinstance(node, ast.BoolOp):
        return find_multiple_node_references(node.values)
    elif isinstance(node, ast.Compare):
        # Interestingly, this is not a BoolOp.
        return find_multiple_node_references([node.left] + node.comparators)
    elif isinstance(node, ast.Call):
        return find_multiple_node_references(
            [node.func] + list(node.args) + [keyword.value for keyword in node.keywords]
        )
    elif isinstance(node, ast.Starred):
        return find_references(node.value)
    elif isinstance(node, ast.IfExp):
        return find_multiple_node_references([node.test, node.body, node.orelse])
    elif isinstance(node, ast.Attribute):
        # Check if this is a simple a.b.c reference. This does some
        # duplicate work with long attribute chains, but it is trivial
        # in the context of a CLI call.
        unwrap: ast.expr = node
        names = []
        while isinstance(unwrap, ast.Attribute):
            names.append(unwrap.attr)
            unwrap = unwrap.value
        if isinstance(unwrap, ast.Name):
            names.append(unwrap.id)
            return set(), {tuple(reversed(names))}
        # It is not really possible to do unusual references with the
        # `attr`; the ast of `a.b(...)` looks like Call(Attribute(...,
        # attr='b'), args=...). It is not possible to do `a.(b+c)`,
        # and `(a+b).b.c` should be automatically handled.
        return find_references(node.value)
    elif isinstance(node, ast.Subscript):
        return find_multiple_node_references([node.value, node.slice])
    elif isinstance(node, ast.Slice):
        nodes = []
        if node.lower:
            nodes.append(node.lower)
        if node.upper:
            nodes.append(node.upper)
        if node.step:
            nodes.append(node.step)
        return find_multiple_node_references(nodes)
    elif isinstance(node, ast.Raise):
        nodes = []
        if node.exc:
            nodes.append(node.exc)
        if node.cause:
            nodes.append(node.cause)
        return find_multiple_node_references(nodes)
    elif isinstance(node, ast.Delete):
        # We don't want to disallow `del`, since __delattr__ might do
        # something significant.
        return find_multiple_node_references(node.targets)
    elif isinstance(node, ast.If) or isinstance(node, ast.While):
        check_nodes = [node.test] + node.body + node.orelse
        return find_multiple_node_references(check_nodes)
    elif isinstance(node, ast.Try) or (
        sys.version_info >= (3, 11, 0) and isinstance(node, ast.TryStar)  # type: ignore
    ):
        try_nodes = node.body + node.handlers + node.orelse + node.finalbody
        return find_multiple_node_references(try_nodes)
    elif (
        isinstance(node, ast.Return)
        or isinstance(node, ast.Yield)
        or isinstance(node, ast.YieldFrom)
    ):
        if node.value:
            return find_references(node.value)
        else:
            return set(), set()
    elif isinstance(node, ast.Await):
        return find_references(node.value)
    # Handle syntax that temporarily binds values.
    elif (
        isinstance(node, ast.ListComp)
        or isinstance(node, ast.SetComp)
        or isinstance(node, ast.GeneratorExp)
        or isinstance(node, ast.DictComp)
    ):
        local_binds = set()
        bound_vars = set()
        refs = set()
        # Earlier generators can be used in later generators.
        for gen in node.generators:
            target_binds, target_refs = find_assignment_lhs_references([gen.target])
            local_binds.update(target_binds)
            # Inline assignments cannot be used in iterables, so we
            # don't have to worry about a more permanent bind
            # overwriting the temporary one.
            iter_binds, iter_refs = find_references(gen.iter)
            assert len(iter_binds) == 0
            if_binds, if_refs = find_multiple_node_references(gen.ifs)
            bound_vars.update(if_binds)
            refs.update(
                var_base_difference(target_refs | iter_refs | if_refs, local_binds)
            )
        elt_binds, elt_refs = (
            find_multiple_node_references([node.key, node.value])
            if isinstance(node, ast.DictComp)
            else find_references(node.elt)
        )
        bound_vars.update(elt_binds)
        refs.update(var_base_difference(elt_refs, local_binds))
        return bound_vars, refs
    elif isinstance(node, ast.ExceptHandler):
        bound_vars, refs = find_multiple_node_references(node.body)
        if node.name is not None:
            refs = var_base_difference(refs, {node.name})
        return bound_vars, refs
    elif isinstance(node, ast.With) or isinstance(node, ast.AsyncWith):
        binds, refs = find_multiple_node_references(node.body)
        # Assignment via with persists beyond the block.
        for item in node.items:
            assert isinstance(item, ast.withitem)
            if item.optional_vars:
                as_binds, as_refs = find_assignment_lhs_references([item.optional_vars])
                binds.update(as_binds)
                refs.update(as_refs)
            ctx_binds, ctx_refs = find_references(item.context_expr)
            binds.update(ctx_binds)
            refs.update(ctx_refs)
        return binds, refs
    elif sys.version_info >= (3, 10, 0) and isinstance(node, ast.Match):  # type: ignore
        # Extracted components are persistent.
        match_nodes = [node.subject] + node.cases
        return find_multiple_node_references(match_nodes)
    elif sys.version_info >= (3, 10, 0) and isinstance(node, ast.match_case):
        case_binds, case_refs = find_match_case_references(node.pattern)
        match_case_nodes: list[ast.AST] = []
        if node.guard:
            match_case_nodes.append(node.guard)
        match_case_nodes.extend(node.body)
        binds, refs = find_multiple_node_references(match_case_nodes)
        return binds | case_binds, refs | case_refs
    # Handle actual binds.
    elif (
        isinstance(node, ast.Assign)
        or isinstance(node, ast.AnnAssign)
        or isinstance(node, ast.AugAssign)
    ):
        targets = []
        if isinstance(node, ast.Assign):
            targets = node.targets
        else:
            targets = [node.target]
        bound_vars, lhs_refs = find_assignment_lhs_references(targets)
        # Type annotated assignments can omit an actual assignment.
        # We could treat this var as free, or raise an error, but
        # instead of trying to guess the user's intent we just let
        # them deal with the runtime consequences of whatever they're
        # up to.
        if hasattr(node, "value") and node.value:
            rhs_bound_vars, rhs_refs = find_references(node.value)
            return bound_vars | rhs_bound_vars, lhs_refs | rhs_refs
        else:
            return bound_vars, lhs_refs
    elif isinstance(node, ast.NamedExpr):
        # Example syntax: (x := 4) (walrus operator).  Walrus
        # disallowed with attributes, subscripts, or tuple unpacking
        # (as of 3.11.2), but treat it as a general assignment LHS,
        # just in case.
        lhs_bound_vars, lhs_refs = find_assignment_lhs_references([node.target])
        rhs_bound_vars, rhs_refs = find_references(node.value)
        return lhs_bound_vars | rhs_bound_vars, lhs_refs | rhs_refs
    elif isinstance(node, ast.Name):
        # If we see a Name, and it isn't otherwise bound, it must be free.
        return (set(), {(node.id,)})
    elif isinstance(node, ast.Import) or isinstance(node, ast.ImportFrom):
        return find_multiple_node_references(node.names)
    elif isinstance(node, ast.alias):
        if node.asname:
            return {node.asname}, set()
        else:
            # This can lead to weird behavior: `import xml` will mask
            # any reference to xml.etree.ElementTree and prevent it
            # from being auto-imported.  Trying to work around this
            # seems absurdly fiddly for a weird corner case, so I'm
            # not going to do anything special about this.
            # Note that the parser automatically handles relative
            # imports so we don't need to do anything with munging
            # away the leading period.
            return {node.name.split(".")[0]}, set()
    elif isinstance(node, ast.For) or isinstance(node, ast.AsyncFor):
        # Unlike generators, `for` bindings stick around afterwards,
        # so definitely don't try to filter them out.
        # You can do things like `for math.x in range(2):`, so re-use
        # normal assignment logic.
        target_bindings, target_refs = find_assignment_lhs_references([node.target])
        for_nodes = [node.iter] + node.body + node.orelse
        body_bindings, body_refs = find_multiple_node_references(for_nodes)
        return target_bindings | body_bindings, target_refs | body_refs
    elif isinstance(node, ast.FunctionDef) or isinstance(node, ast.AsyncFunctionDef):
        fn_nodes = [node.args] + node.body + node.decorator_list
        bound_vars, refs = find_multiple_node_references(fn_nodes)
        return bound_vars | {node.name}, refs
    elif isinstance(node, ast.Lambda):
        return find_multiple_node_references([node.args, node.body])
    elif isinstance(node, ast.arguments):
        arg_nodes = (
            node.posonlyargs
            + node.args
            + node.kwonlyargs
            + [node.vararg]
            + [node.kwarg]
            + node.kw_defaults
            + node.defaults
        )
        return find_multiple_node_references([n for n in arg_nodes if n])
    elif isinstance(node, ast.arg):
        return {node.arg}, set()
    elif isinstance(node, ast.keyword):
        binds, refs = find_references(node.value)
        return binds | ({node.arg} if node.arg else set()), refs
    elif isinstance(node, ast.ClassDef):
        # Keywords are not actually used anywhere (not real bindings), but the values can have refs.
        class_nodes = node.bases + node.keywords + node.body + node.decorator_list
        bindings, refs = find_multiple_node_references(class_nodes)
        return {node.name} | bindings, refs
    # Explicitly don't do anything.
    elif (
        isinstance(node, ast.Constant)
        or isinstance(node, ast.Pass)
        or (sys.version_info >= (3, 12, 0) and isinstance(node, ast.TypeAlias))  # type: ignore
        or isinstance(node, ast.Break)
        or isinstance(node, ast.Continue)
        or
        # The referred variables aren't actually assigned this way.
        isinstance(node, ast.Global)
        or isinstance(node, ast.Nonlocal)
    ):
        return set(), set()
    # Fallback case.
    else:
        LOG.warning("Skipping reference checking for {}".format(node))
        return (set(), set())


def find_multiple_node_references(
    nodes: Sequence[ast.AST],
) -> tuple[set[str], set[tuple[str, ...]]]:
    """Helper function to gather all reference types from a list of nodes."""
    bound_vars = set()
    refs = set()
    for node in nodes:
        node_bound_vars, node_refs = find_references(node)
        bound_vars.update(node_bound_vars)
        refs.update(node_refs)
    return bound_vars, refs


def find_assignment_lhs_references(
    targets: Sequence[ast.expr],
) -> tuple[set[str], set[tuple[str, ...]]]:
    """Only a subset of ast nodes are allowed on the LHS of an
    assignment, and names are bindings instead of free variables.
    """
    bound_vars = set()
    lhs_refs = set()
    for t in targets:
        if isinstance(t, ast.Name):
            bound_vars.add(t.id)
        elif isinstance(t, ast.Tuple) or isinstance(t, ast.List):
            tmp_binds, tmp_refs = find_assignment_lhs_references(t.elts)
            bound_vars.update(tmp_binds)
            lhs_refs.update(tmp_refs)
        else:
            # Anything that is not a plain name is actually a reference.
            # Think `module.flag = 1`.
            tmp_bound_vars, tmp_refs = find_references(t)
            # Defend against := showing up in the LHS.
            assert len(tmp_bound_vars) == 0
            lhs_refs.update(tmp_refs)
    return bound_vars, lhs_refs


if sys.version_info >= (3, 10, 0):

    def find_match_case_references(
        target: ast.pattern,
    ) -> tuple[set[str], set[tuple[str, ...]]]:
        """Find variables assignments within match case statements."""
        binds = set()
        refs = set()
        if isinstance(target, ast.MatchValue):
            # This can match attributes like `x.a`, so we should use the
            # usual LHS rules.
            value_binds, value_refs = find_assignment_lhs_references([target.value])
            binds |= value_binds
            refs |= value_refs
        elif isinstance(target, ast.MatchSingleton):
            # Used for True/False/None.
            pass
        elif isinstance(target, ast.MatchSequence):
            find_multiple_match_case_references(target.patterns, binds, refs)
        elif isinstance(target, ast.MatchMapping):
            key_binds, key_refs = find_assignment_lhs_references(target.keys)
            binds |= key_binds
            refs |= key_refs
            find_multiple_match_case_references(target.patterns, binds, refs)
            if target.rest:
                binds.add(target.rest)
        elif isinstance(target, ast.MatchClass):
            cls_binds, class_refs = find_references(target.cls)
            binds |= cls_binds
            refs |= class_refs
            find_multiple_match_case_references(target.patterns, binds, refs)
            if target.kwd_patterns:
                find_multiple_match_case_references(target.kwd_patterns, binds, refs)
        elif isinstance(target, ast.MatchStar):
            if target.name:
                binds.add(target.name)
        elif isinstance(target, ast.MatchAs):
            if target.name:
                binds.add(target.name)
            if target.pattern:
                as_binds, as_refs = find_match_case_references(target.pattern)
                binds |= as_binds
                refs |= as_refs
        elif isinstance(target, ast.MatchOr):
            find_multiple_match_case_references(target.patterns, binds, refs)
        return binds, refs

    def find_multiple_match_case_references(
        patterns: list[ast.pattern], binds: set[str], refs: set[tuple[str, ...]]
    ):
        # We normally don't modify parameters, but the match semantics are
        # simple enough we can get away with it this time.
        for pattern in patterns:
            tmp_binds, tmp_refs = find_match_case_references(pattern)
            binds |= tmp_binds
            refs |= tmp_refs



#  Copyright (c) <2014> <thenoviceoof>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

import logging
import sys
from pyli.main import main

__version__ = (2, 0, 1)

# Any sane person would use argparse; however, we want to accept
# arbitrary switches, so unfortunately argparse is not an option.

HELP_MSG = """
pyli is a utility to make using python in conjunction with other CLI tools easier
 - attempts to auto import unbound variables
 - populate special variables (lines, line, contents) with structured
   data from stdin
 - print the last line automatically (if not None)
 - provides command line options as variables (other than those listed
   below)

Special switches include:
 -v, -vv, --debug  Outputs debug information useful when developing pyli.
 --help            Outputs this message.
 -pp, --pprint     Uses pprint.pprint() instead of python's builtin print.
 --version         Outputs the current version of pyli.

Check out https://github.com/thenoviceoof/pyli for more details!
"""


# TODO: add a --debug-out switch to provide an alternative for debug
# info than stderr.
def script_entry_point():
    if len(sys.argv) == 1 or "--help" in sys.argv:
        print(HELP_MSG.format(sys.argv[0]))
    elif "--version" in sys.argv:
        version_string = ".".join(str(v) for v in __version__)
        print(version_string)
    else:
        args = sys.argv[1:]
        debug = logging.ERROR
        pprint = False
        # strip out any switches
        if "-v" in args:
            args.remove("-v")
            debug = logging.WARNING
        if "-vv" in args:
            args.remove("-vv")
            debug = logging.INFO
        if "--debug" in args:
            args.remove("--debug")
            debug = logging.DEBUG
        if "--pprint" in args:
            args.remove("--pprint")
            pprint = True
        if "-pp" in args:
            args.remove("-pp")
            pprint = True
        # pass everything else as a variable
        commands = []
        kwargs: dict[str, str | bool] = {}
        while args:
            if args[0][0] == "-":
                if "=" in args[0]:
                    name, val = args[0].split("=", 1)
                    name = name.lstrip("-")
                    kwargs[name] = val
                    args = args[1:]
                elif len(args) > 1 and args[1][0] != "-":
                    name, val = args[:2]
                    name = name.lstrip("-")
                    kwargs[name] = val
                    args = args[2:]
                else:
                    # treat as a boolean switch
                    name = args[0].strip("-")
                    kwargs[name] = True
                    args = args[1:]
            else:
                commands.append(args[0])
                args = args[1:]

        program = "\n".join(commands)
        main(program, debug=debug, pprint_opt=pprint, variables=kwargs)



#  Copyright (c) <2014> <thenoviceoof>
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

import ast
import logging
from collections.abc import Sequence

LOG = logging.getLogger(__name__)


def create_imports(tree: ast.AST, free_variables: set[tuple[str, ...]]) -> None:
    LOG.info("Creating imports...")
    assert isinstance(tree, ast.Module)
    for free_var in free_variables:
        LOG.debug("Creating import chain for {}".format(".".join(free_var)))
        # Build up a try-import chain from most specific to least.
        current_import: ast.stmt = create_import(free_var[:1])
        for i in range(2, len(free_var) + 1):
            current_import = ast.Try(
                body=[create_import(free_var[:i])],
                handlers=[
                    ast.ExceptHandler(
                        type=ast.Name(id="ImportError", ctx=ast.Load()),
                        body=[current_import],
                    )
                ],
                orelse=[],
                finalbody=[],
            )
        # This should continually copy line 0 to each new import.
        ast.copy_location(current_import, tree.body[0])
        # It is possible to keep pushing the rest of the body down,
        # but for some reason this easily ends up with impossible line
        # ranges (start larger than end).
        tree.body.insert(0, current_import)


def create_import(import_path: Sequence[str]) -> ast.Import:
    return ast.Import(names=[ast.alias(name=".".join(import_path))])
